
 {***************************************************************************
                          REALISATION DU KHOLLOSCOPE
                                 Denis Cazor

                            denis.cazor@wanadoo.fr
 ***************************************************************************}

 {$m $4000, 0, $8000}

 PROGRAM clscp ;

 USES    crt, dos ;

 CONST   max_matieres = 11 ;                 { nombre de matires  traiter }
         max_semaines = 72 ;               { nombre de semaines  optimiser }
         max_heures   =  9 ;   { nombre d'heures de colles par matire < 10 }
         max_blocs    =  9 ;               { nombre de blocs  traiter < 10 }
         max_conflits = 60 ;               { nombre de conflits  envisager }
         max_erreur   = 10 ;                    { nombre de types d'erreurs }
         max_lignes   = 24 ;                 { nombre de semaines affiches }
         max_buffer   = 12 ;          { tampon d'enregistrement des groupes }

         mode_emploi : ARRAY[0..3] OF STRING[80]
            = ('F1234 remplis F5678 optimise F9 mmorise Fa rapelle Fb tri',
               'A-F1234 reset limite A-F5 tore A-F6 curseur',
               'A-F9 enregistre  A-Fa recharge', '   A-F8 reset') ;

         messages_erreur : ARRAY[0..max_erreur] OF STRING[20]
            = ('', 'Groupe Identiques', 'Bloc Identiques', 'Groupe invalide',
               'Collision colles', 'Collision TDs', 'Contraintes',
               'Amnagements TDs', 'Amnagements Sup', 'Horaire incorrect',
               'Heures conscutives') ;

         poids_erreurs : ARRAY[0..max_erreur] OF WORD
            = (0, 23, 23, 1, 71, 71, 23, 7, 7, 223, 71) ;

         permutation_size = 5 ;                  { tendue des permutations }

         lineset_on  = $0004 ;
         lineset_off = $0800 ;
              { gestion du curseur ON/OFF - spcifique  la carte graphique }
         ret = #$a + #$d ;
         esc = #27 ;

 TYPE    data = RECORD
                  erreur : WORD ;
                  groupe : CHAR ;
               END ;

         type_cellule = RECORD
                  heure, semaine : WORD ;
                  matiere : CHAR ;
               END ;

         type_bloc = RECORD
                  semaine : ARRAY[1..4] OF WORD ;
                  periode, long : WORD ;
                  matiere : STRING[4] ;
               END ;

         groupes_bloc = ARRAY[1..4 * max_heures] OF CHAR ;

         colloscope = ARRAY['a'..chr(max_matieres + ord('a') - 1),
               1..max_semaines, 1..max_heures] OF data ;

 VAR     collisions_tds, conflits_colles, contraintes_colles,
            successions_colles : ARRAY[1..max_conflits] OF STRING[4] ;
         obligations_langues : ARRAY[1..max_conflits] OF STRING[20] ;
         liste_symboles, result_file, reference_file : STRING[120] ;
         amenagements_sup, amenagements_tds,
            definition_blocs : ARRAY[1..max_blocs] OF STRING[4] ;
         sequence_optimisation : STRING[max_conflits] ;
         liste_colles, liste_dispersion, liste_regular,
            liste_tds : STRING[max_blocs] ;
         dernier_groupe, derniere_matiere : CHAR ;
         titres : ARRAY[1..2] OF STRING[80] ;

         matrice : ARRAY[1..max_semaines, 1..max_semaines] OF WORD ;
         data_chi2 : ARRAY[1..max_heures, 'A'..'Z'] OF INTEGER ;
         liste_blocs : ARRAY[1..max_blocs] OF type_bloc ;
         subs_group : ARRAY[0..max_buffer] OF CHAR ;
         position : ARRAY[1..max_semaines] OF WORD ;
         color : ARRAY[1..max_semaines] OF BOOLEAN ;
         erreur_groupes : ARRAY['A'..'W'] OF WORD ;

         cycle_eff, cycle_min, cycle_max, difference, nb_heures,
            duree_optimale, max_amenagements, max_periode, min_periode,
         nb_amenagements_sup, nb_amenagements_tds, nb_blocs, nb_reussites,
            nb_conflits_colles, nb_collisions_tds, nb_contraintes_colles,
         nb_elements, nb_erreurs, nb_groupes, nb_lignes, sous_diagonale,
            nb_matieres, nb_obligations_langues, old_nber, premiere,
         nb_semaines, nb_successions_colles, semaines_effectives,
            limite_optimisation, nb_colles : WORD ;
         amenagements, chi2_limite, chi2_reduit, mode_torique, mode_curseur,
            modified, working : BOOLEAN ;

         memoire_tab : ^colloscope ;
         nb_similitudes : DOUBLE ;
         curseur : type_cellule ;
         chi2_cumul : INTEGER ;
         tableau : colloscope ;
         nb_tests : LONGINT ;

 {***************************************************************************
                Ralisation du colloscope - Fonctions de base
 ***************************************************************************}

 PROCEDURE arret(message : STRING) ;

 BEGIN
   textbackground(lightgray) ;
   writeln(ret, '  ', message, ' !', ret) ;
   readkey ;
   halt ;
 END ; { arret }


 PROCEDURE beep ;

 BEGIN
   nosound ;
   sound(440) ;
   delay(100) ;
   nosound ;
 END ; { beep }


 FUNCTION readkey : CHAR ;

 VAR     debut : WORD ABSOLUTE $40:$1a ;
         fin : WORD ABSOLUTE $40:$1c ;
         adresse : ^WORD ;
         data : WORD ;

 BEGIN
   REPEAT UNTIL fin <> debut ;
   adresse := ptr($40, debut) ;
   data := adresse^ ;
   IF (lo(data) = 0) OR (lo(data) = $e0) OR odd(debut) THEN
      BEGIN
         inc(debut) ;
         IF debut > $3d THEN debut := $1e ;
         IF lo(data) = $e0 THEN readkey := #0
            ELSE readkey := chr(lo(data)) ;
      END
   ELSE
      BEGIN
         inc(debut, 2) ;
         IF debut > $3d THEN debut := $1e ;
         readkey := chr(lo(data)) ;
      END ;
 END ; { readkey }


 PROCEDURE cursor(lineset : WORD) ;

 VAR     regs : REGISTERS ;

 BEGIN
   regs.ax := $100 ;
   regs.cx := lineset ;
   intr(16, regs) ;
 END ; { cursor }


 FUNCTION existe(nomfich : STRING) : BOOLEAN ;

 VAR     ioerr : BOOLEAN ;
         fichier : FILE ;

 BEGIN
   ioerr := (ioresult = 0) ;
   {$i-}
   assign(fichier, nomfich) ;
   filemode := 0 ;
   reset(fichier) ;
   ioerr := (ioresult = 0) ;
   IF ioerr THEN close(fichier) ;
   {$i+}
   existe := ioerr ;
 END ; { existe }


 FUNCTION inf(a, b : LONGINT) : LONGINT ;

 BEGIN
   IF a < b THEN inf := a ELSE inf := b ;
 END ; { inf }


 FUNCTION pgcd(a, b : LONGINT) : LONGINT ;

 BEGIN
   IF b = 0 THEN pgcd := a
      ELSE pgcd := pgcd(b, a MOD b) ;
 END ; { pgcd }


 FUNCTION ppcm(a, b : LONGINT) : LONGINT ;

 BEGIN
   ppcm := (a * b) DIV pgcd(a, b) ;
 END ; { ppcm }


 FUNCTION ln_fact(n : WORD) : EXTENDED ;

 VAR     valeur : EXTENDED ;
         int : WORD ;

 BEGIN
   valeur := 0 ;
   FOR int := 2 TO n DO valeur :=  valeur + ln(int) ;
   ln_fact := valeur ;
 END ; { ln_fact }


 FUNCTION str_hex(num : LONGINT ; size : BYTE) : STRING ;

 VAR     chaine : STRING[10] ;
         modulo : BYTE ;

 BEGIN
   chaine := '' ;
   REPEAT
      modulo := num AND $f ;
      num := num SHR 4 ;
      IF modulo > 9
         THEN chaine := CHAR(modulo + ord('A') - 10) + chaine
         ELSE chaine := CHAR(modulo + ord('0')) + chaine ;
   UNTIL num = 0 ;
   WHILE length(chaine) < size DO chaine := ' ' + chaine ;
   str_hex := chaine ;
 END ; { str_hex }


 FUNCTION recherche_bloc(matiere : CHAR) : WORD ;

 VAR     ix : WORD ;

 BEGIN
   ix := 1 ;
   WHILE pos(matiere, definition_blocs[ix]) = 0 DO inc(ix) ;
   recherche_bloc := ix ;
 END ; { recherche_bloc }


 FUNCTION exclusion(a : STRING) : BOOLEAN ;

 VAR     result : BOOLEAN ;
         ix : WORD ;

 BEGIN
   result := true ;
   FOR ix := 1 TO length(a) DO
      result := result AND (pos(a[ix], liste_symboles) > 0) ;
   exclusion := NOT result ;
 END ; { exclusion }

 {***************************************************************************
           Fonctions de gnration des permutations et combinaisons
 ***************************************************************************}

 FUNCTION compatible(i, p : WORD) : BOOLEAN ;

 VAR     acceptable : BOOLEAN ;

 BEGIN
   acceptable := true ;
   WHILE acceptable AND (0 < i) DO
      BEGIN
         acceptable := (p <> position[i]) ;
         dec(i) ;
      END ;
   compatible := acceptable ;
 END ; { compatible }


 PROCEDURE next_permutation ;       { gnration de la permutation suivante }

 VAR     i, j : WORD ;

 BEGIN
   j := nb_elements ;
   i := j - 1 ;
   REPEAT
      inc(position[j]) ;
      IF position[j] > nb_elements THEN
         BEGIN
            position[j] := 0 ;
            dec(i) ;
            dec(j) ;
         END
      ELSE IF compatible(i, position[j]) THEN
         BEGIN
            inc(i) ;
            inc(j) ;
         END ;
   UNTIL (j > nb_elements) OR (j = 0) ;
   difference := 0 ;
   FOR i := 1 TO nb_elements DO IF position[i] <> i THEN inc(difference) ;
 END ; { next_permutation }


 PROCEDURE new_permutation ;        { gnration d'une nouvelle permutation }

 VAR     i, j, temp : WORD ;

 BEGIN
   i := random(cycle_eff) ;
   REPEAT
      j := random(cycle_eff) ;
   UNTIL i <> j ;
   inc(i) ;
   inc(j) ;
   temp := position[i] ;
   position[i] := position[j] ;
   position[j] := temp ;
 END ; { new_permutation }


 PROCEDURE next_combinaison ;       { gnration de la combinaison suivante }

 VAR     i, j : WORD ;

 BEGIN
   i := 1 ;
   j := 2 ;
   WHILE (j <= nb_elements) AND (position[i] + 1 = position[j]) DO
      BEGIN
         inc(i) ;
         inc(j) ;
      END ;
   inc(position[i]) ;
   FOR j := 1 TO i - 1 DO position[j] := j ;
 END ; { next_combinaison }

 {***************************************************************************
                  Ralisation du colloscope - Srie de tests
 ***************************************************************************}

 PROCEDURE enregistre_erreur(matiere : CHAR ; heure, semaine, numero : WORD) ;

 BEGIN
   IF (poids_erreurs[tableau[matiere, semaine, heure].erreur] <
   poids_erreurs[numero]) AND (tableau[matiere, semaine, heure].groupe < 'X')
         THEN tableau[matiere, semaine, heure].erreur := numero ;
 END ; { enregistre_erreur }


 PROCEDURE test_gd(semaine : WORD) ;        { les groupes sont distincts }

 VAR     heure, heurp : WORD ;
         matiere : CHAR ;

 BEGIN
   FOR matiere := 'a' TO derniere_matiere DO
      FOR heure := 1 TO nb_heures - 1 DO
         FOR heurp := heure + 1 TO nb_heures DO
      BEGIN
         inc(nb_tests) ;
         IF (tableau[matiere, semaine, heure].groupe =
               tableau[matiere, semaine, heurp].groupe) THEN
            BEGIN
               enregistre_erreur(matiere, heure, semaine, 1) ;
               enregistre_erreur(matiere, heurp, semaine, 1) ;
            END ;
      END ;
 END ; { test_gd }


 PROCEDURE test_blocs(semaine : WORD) ;

 VAR     ix, iy, iz, semaine_md, semainp, semainq : WORD ;
         matierp, matierq : CHAR ;

 PROCEDURE _test1 ;

 VAR     heurp, heurq : WORD ;

 BEGIN
   FOR heurp := 1 TO nb_heures DO
         FOR heurq := 1 TO nb_heures DO
      BEGIN
         inc(nb_tests) ;
         IF (tableau[matierp, semainp, heurp].groupe =
               tableau[matierq, semainq, heurq].groupe) THEN
            BEGIN
               enregistre_erreur(matierp, heurp, semainp, 2) ;
               enregistre_erreur(matierq, heurq, semainq, 2) ;
            END ;
      END ;
 END ; { _test1 }

 BEGIN
   FOR ix := 1 TO nb_blocs DO
      BEGIN
         semaine_md := semaine - (semaine - 1) MOD liste_blocs[ix].periode ;
         FOR iy := 1 TO liste_blocs[ix].long - 1 DO
               FOR iz := iy + 1 TO liste_blocs[ix].long DO
            BEGIN
               matierp := liste_blocs[ix].matiere[iy] ;
               semainp := semaine_md + liste_blocs[ix].semaine[iy] ;
               matierq := liste_blocs[ix].matiere[iz] ;
               semainq := semaine_md + liste_blocs[ix].semaine[iz] ;
               _test1 ;
            END ;
      END ;
 END ; { test_blocs }


 PROCEDURE test_gc(semaine : WORD) ;
                                     { les numros de groupes sont corrects }
 VAR     matiere : CHAR ;
         heure : WORD ;

 BEGIN
   FOR matiere := 'a' TO derniere_matiere DO
         FOR heure := 1 TO nb_heures DO
      BEGIN
         inc(nb_tests) ;
         IF (ord(tableau[matiere, semaine, heure].groupe) >= ord('A') +
            nb_groupes) THEN enregistre_erreur(matiere, heure, semaine, 3) ;
      END ;
 END ; { test_gc }


 PROCEDURE tests_conflits_colles(semaine : WORD) ;
              { test heure / heure - les heures de passages sont distinctes }
 VAR     num, heure, heurp : WORD ;
         matiere, matierp : CHAR ;

 PROCEDURE _test2 ;

 BEGIN
   inc(nb_tests) ;
   IF (tableau[matiere, semaine, heure].groupe =
         tableau[matierp, semaine, heurp].groupe) THEN
      BEGIN
         enregistre_erreur(matiere, heure, semaine, 4) ;
         enregistre_erreur(matierp, heurp, semaine, 4) ;
      END ;
 END ; { _test2 }

 BEGIN
   FOR num := 1 TO nb_conflits_colles DO
      BEGIN
         matiere := conflits_colles[num][1] ;
         heure := ord(conflits_colles[num][2]) - ord('0') ;
         matierp := conflits_colles[num][3] ;
         heurp := ord(conflits_colles[num][4]) - ord('0') ;
         _test2 ;
      END ;
 END ; { tests_conflits_colles }


 PROCEDURE tests_successions_colles(semaine : WORD) ;
              { test heure / heure - les heures de passages sont distinctes }
 VAR     num, heure, heurp : WORD ;
         matiere, matierp : CHAR ;

 PROCEDURE _test2 ;

 BEGIN
   inc(nb_tests) ;
   IF (tableau[matiere, semaine, heure].groupe =
         tableau[matierp, semaine, heurp].groupe) THEN
      BEGIN
         enregistre_erreur(matiere, heure, semaine, 10) ;
         enregistre_erreur(matierp, heurp, semaine, 10) ;
      END ;
 END ; { _test2 }

 BEGIN
   FOR num := 1 TO nb_successions_colles DO
      BEGIN
         matiere := successions_colles[num][1] ;
         heure := ord(successions_colles[num][2]) - ord('0') ;
         matierp := successions_colles[num][3] ;
         heurp := ord(successions_colles[num][4]) - ord('0') ;
         _test2 ;
      END ;
 END ; { tests_successions_colles }


 PROCEDURE tests_collisions_tds(semaine : WORD) ;
            { test heure / matiere - les heures de passages sont distinctes }
 VAR     matiere, matierp : CHAR ;
         num, heure : WORD ;

 PROCEDURE _test4 ;

 VAR     heurp : WORD ;

 BEGIN
   FOR heurp := 1 TO nb_heures DO
      BEGIN
         inc(nb_tests) ;
         IF (tableau[matiere, semaine, heure].groupe =
               tableau[matierp, semaine, heurp].groupe) THEN
            BEGIN
               enregistre_erreur(matiere, heure, semaine, 5) ;
               enregistre_erreur(matierp, heurp, semaine, 5) ;
            END ;
      END ;
 END ; { _test4 }

 BEGIN
   FOR num := 1 TO nb_collisions_tds DO
      BEGIN
         matiere := collisions_tds[num][1] ;
         heure := ord(collisions_tds[num][2]) - ord('0') ;
         matierp := collisions_tds[num][3] ;
         _test4 ;
      END ;
 END ; { tests_collisions_tds }


 PROCEDURE tests_contraintes(semaine : WORD) ;
        { test matiere / matiere - les semaines de passages sont distinctes }
 VAR     matiere, matierp : CHAR ;
         num : WORD ;

 PROCEDURE _test5 ;

 VAR     heure, heurp : WORD ;

 BEGIN
   FOR heure := 1 TO nb_heures DO
         FOR heurp := 1 TO nb_heures DO
      BEGIN
         inc(nb_tests) ;
         IF (tableau[matiere, semaine, heure].groupe =
                  tableau[matierp, semaine, heurp].groupe) THEN
            BEGIN
               enregistre_erreur(matiere, heure, semaine, 6) ;
               enregistre_erreur(matierp, heurp, semaine, 6) ;
            END ;
      END ;
 END ; { _test5 }

 BEGIN
   FOR num := 1 TO nb_contraintes_colles DO
      BEGIN
         matiere := contraintes_colles[num][1] ;
         matierp := contraintes_colles[num][2] ;
         _test5 ;
      END ;
 END ; { tests_contraintes }


 PROCEDURE tests_amenagements_tds(semaine : WORD) ;
            { test heure / matiere - les heures de passages sont distinctes }
 VAR     matiere, matierp : CHAR ;
         num, heure : WORD ;

 PROCEDURE _test3 ;

 VAR     heurp : WORD ;

 BEGIN
   FOR heurp := 1 TO nb_heures DO
      BEGIN
         inc(nb_tests) ;
         IF (tableau[matiere, semaine, heure].groupe =
               tableau[matierp, semaine, heurp].groupe) THEN
            BEGIN
               enregistre_erreur(matiere, heure, semaine, 7) ;
               enregistre_erreur(matierp, heurp, semaine, 7) ;
            END ;
      END ;
 END ; { _test3 }

 BEGIN
   IF max_amenagements = 0 THEN exit ;
   FOR num := 1 TO inf(max_amenagements, nb_amenagements_tds) DO
      BEGIN
         matiere := amenagements_tds[num][1] ;
         heure := ord(amenagements_tds[num][2]) - ord('0') ;
         matierp := amenagements_tds[num][3] ;
         _test3 ;
      END ;
 END ; { tests_amenagements_tds }


 PROCEDURE tests_amenagements_sup(semaine : WORD) ;
        { test matiere / matiere - les semaines de passages sont distinctes }
 VAR     matiere, matierp : CHAR ;
         num : WORD ;

 PROCEDURE _test6 ;

 VAR     heure, heurp : WORD ;

 BEGIN
   FOR heure := 1 TO nb_heures DO
         FOR heurp := 1 TO nb_heures DO
      BEGIN
         inc(nb_tests) ;
         IF (tableau[matiere, semaine, heure].groupe =
               tableau[matierp, semaine, heurp].groupe) THEN
            BEGIN
               enregistre_erreur(matiere, heure, semaine, 8) ;
               enregistre_erreur(matierp, heurp, semaine, 8) ;
            END ;
      END ;
 END ; { _test6 }

 BEGIN
   IF max_amenagements <= nb_amenagements_tds THEN exit ;
   FOR num := 1 TO max_amenagements - nb_amenagements_tds DO
      BEGIN
         matiere := amenagements_sup[num][1] ;
         matierp := amenagements_sup[num][2] ;
         _test6 ;
      END ;
 END ; { tests_amenagements_sup }


 PROCEDURE test_langues(semaine : WORD) ;
             { tests heures / groupes - seuls les groupes concerns passent }
 VAR     heure, num : WORD ;
         groupes : STRING ;
         nature : BOOLEAN ;
         matiere : CHAR ;

 PROCEDURE _test_lv ;

 BEGIN
   inc(nb_tests, length(groupes)) ;
   IF (pos(tableau[matiere, semaine, heure].groupe, groupes) = 0)
      xor nature THEN enregistre_erreur(matiere, heure, semaine, 9) ;
 END ; { _test_lv }

 BEGIN
   FOR num := 1 TO nb_obligations_langues DO
      BEGIN
         nature := (obligations_langues[num][1] = '-') ;
         matiere := obligations_langues[num][2] ;
         heure := ord(obligations_langues[num][3]) - ord('0') ;
         groupes := copy(obligations_langues[num], 4, 55) ;
         _test_lv ;
      END ;
 END ; { test_langues }

 {***************************************************************************
                        Srie de tests - Bilan complet
 ***************************************************************************}

 FUNCTION decompte_erreurs(semaine : WORD) : WORD ;

 VAR     heure, nb_errs, semainp : WORD ;
         matiere : CHAR ;

 BEGIN
   nb_errs := 0 ;
   dec(semaine, (semaine - 1) MOD max_periode) ;
   FOR matiere := 'a' TO derniere_matiere DO
         FOR heure := 1 TO nb_heures DO
      FOR semainp := semaine TO semaine + min_periode DO
         IF tableau[matiere, semainp, heure].groupe <= dernier_groupe THEN
      inc(nb_errs, poids_erreurs[tableau[matiere, semainp, heure].erreur]) ;
   FOR semainp := semaine TO semaine + min_periode DO
      color[semainp] := (nb_errs > 0) ;
   decompte_erreurs := nb_errs ;
 END ; { decompte_erreurs }


 PROCEDURE test_optimisation(semaine : WORD) ;

 VAR     matiere : CHAR ;
         heure : WORD ;

 BEGIN
   FOR matiere := 'a' TO derniere_matiere DO
         FOR heure := 1 TO nb_heures DO
      tableau[matiere, semaine, heure].erreur := 0 ;
   test_langues(semaine) ;
   tests_contraintes(semaine) ;
   tests_collisions_tds(semaine) ;
   tests_amenagements_tds(semaine) ;
   tests_amenagements_sup(semaine) ;
   tests_conflits_colles(semaine) ;
   tests_successions_colles(semaine) ;
 END ; { test_optimisation }


 PROCEDURE test_global(semaine : WORD) ;

 VAR     matiere : CHAR ;
         heure : WORD ;

 BEGIN
   FOR matiere := 'a' TO derniere_matiere DO
         FOR heure := 1 TO nb_heures DO
      tableau[matiere, semaine, heure].erreur := 0 ;
   test_gc(semaine) ;
   test_gd(semaine) ;
   test_langues(semaine) ;
   test_blocs(semaine) ;
   tests_contraintes(semaine) ;
   tests_collisions_tds(semaine) ;
   tests_amenagements_tds(semaine) ;
   tests_amenagements_sup(semaine) ;
   tests_conflits_colles(semaine) ;
   tests_successions_colles(semaine) ;
 END ; { test_global }


 PROCEDURE evaluation_reduite(semaine : WORD) ;

 VAR     heure, semainp, semaine_ref : WORD ;
         groupe, matiere : CHAR ;

 BEGIN
   nb_tests := 0 ;
   nb_erreurs := 0 ;
   FOR groupe := 'A' TO 'W' DO erreur_groupes[groupe] := 0 ;
   semaine_ref := semaine - (semaine - 1) MOD max_periode ;
   FOR semainp := semaine_ref TO semaine_ref + min_periode DO
         test_global(semainp) ;
   FOR semainp := 1 TO nb_semaines DO
      FOR matiere := 'a' TO derniere_matiere DO
         FOR heure := 1 TO nb_heures DO
      BEGIN
         inc(nb_tests) ;
         IF (tableau[matiere, semainp, heure].erreur > 0) AND
            (tableau[matiere, semainp, heure].groupe <= dernier_groupe) THEN
            BEGIN
               inc(erreur_groupes[tableau[matiere, semainp, heure].groupe]) ;
               inc(nb_erreurs) ;
            END ;
      END ;
 END ; { evaluation_reduite }


 PROCEDURE evalue ;

 VAR     groupe, matiere : CHAR ;
         semaine, heure : WORD ;

 BEGIN
   nb_tests := 0 ;
   nb_erreurs := 0 ;
   FOR groupe := 'A' TO 'W' DO erreur_groupes[groupe] := 0 ;
   FOR semaine := 1 TO nb_semaines DO test_global(semaine) ;
   FOR semaine := 1 TO nb_semaines DO
      FOR matiere := 'a' TO derniere_matiere DO
         FOR heure := 1 TO nb_heures DO
      BEGIN
         inc(nb_tests) ;
         IF (tableau[matiere, semaine, heure].erreur > 0) AND
            (tableau[matiere, semaine, heure].groupe <= dernier_groupe) THEN
            BEGIN
               inc(erreur_groupes[tableau[matiere, semaine, heure].groupe]) ;
               inc(nb_erreurs) ;
            END ;
      END ;
   FOR semaine := 1 TO nb_semaines DO decompte_erreurs(semaine) ;
 END ; { evalue }

 {***************************************************************************
             Ralisation du colloscope - Fonctions de remplissage
 ***************************************************************************}

 PROCEDURE clear_tableau ;

 VAR     semaine, heure : WORD ;
         matiere : CHAR ;

 BEGIN
   FOR semaine := 1 TO nb_semaines DO
         FOR matiere := 'a' TO derniere_matiere DO
      FOR heure := 1 TO nb_heures DO
         BEGIN
            tableau[matiere, semaine, heure].erreur := 0 ;
            tableau[matiere, semaine, heure].groupe := 'Z' ;
         END ;
 END ; { clear_tableau }


 PROCEDURE remplis_premiere(matiere : CHAR) ;

 VAR     ix, semaine, heure, nsb, nsb_ref : WORD ;
         bloc : type_bloc ;
         groupe : CHAR ;

 BEGIN
   modified := true ;
   bloc := liste_blocs[recherche_bloc(matiere)] ;
   FOR ix := 1 TO bloc.long DO inc(bloc.semaine[ix]) ;
   nsb := 1 ;
   WHILE ((bloc.semaine[nsb] <> curseur.semaine) OR (bloc.matiere[nsb] <>
         matiere)) AND (nsb < bloc.long) DO inc(nsb) ;
   nsb_ref := nsb ;
   heure := curseur.heure ;
   groupe := 'A' ;
   REPEAT
      REPEAT
         IF tableau[bloc.matiere[nsb], bloc.semaine[nsb], heure].groupe <>
               'X' THEN
            BEGIN
               tableau[bloc.matiere[nsb],
                     bloc.semaine[nsb], heure].groupe := groupe ;
               groupe := succ(groupe) ;
            END ;
         inc(heure) ;
      UNTIL (heure > nb_heures) OR
            ((nsb = nsb_ref) AND (heure = curseur.heure)) ;
      IF ((nsb <> nsb_ref) OR (heure <> curseur.heure)) THEN
         BEGIN
            heure := 1 ;
            inc(nsb) ;
            IF nsb > bloc.long THEN nsb := 1 ;
         END ;
   UNTIL (nsb = nsb_ref) AND (heure = curseur.heure) ;
 END ; { remplis_premiere }


 PROCEDURE circulation_bloc(bloc : type_bloc) ;

 VAR     groupe, dernier : CHAR ;
         heure, nsb : WORD ;

 BEGIN
   modified := true ;
   dernier := 'A' ;
   FOR nsb := 1 TO bloc.long DO
         FOR heure := 1 TO nb_heures DO
      BEGIN
         groupe := tableau[bloc.matiere[nsb],
                  bloc.semaine[nsb] - bloc.periode, heure].groupe ;
         IF (groupe < 'X') AND (dernier < groupe) THEN dernier := groupe ;
      END ;
   FOR nsb := 1 TO bloc.long DO
         FOR heure := 1 TO nb_heures DO
      BEGIN
         groupe := tableau[bloc.matiere[nsb],
                  bloc.semaine[nsb] - bloc.periode, heure].groupe ;
         CASE groupe OF
            'X' : tableau[bloc.matiere[nsb], bloc.semaine[nsb],
                     heure].groupe := 'X' ;
            'Z' : tableau[bloc.matiere[nsb], bloc.semaine[nsb],
                     heure].groupe := 'Z' ;
            ELSE tableau[bloc.matiere[nsb], bloc.semaine[nsb],
                     heure].groupe := succ(groupe) ;
            END ;
         IF groupe = dernier THEN tableau[bloc.matiere[nsb],
                  bloc.semaine[nsb], heure].groupe := 'A' ;
      END ;
 END ; { circulation_bloc }


 PROCEDURE circulation_sousbloc(bloc : type_bloc) ;

 VAR     heure, heurp, nsb : WORD ;
         groupe : CHAR ;

 BEGIN
   modified := true ;
   FOR nsb := 1 TO bloc.long DO
      BEGIN
         heure := 1 ;
         REPEAT
            groupe := tableau[bloc.matiere[nsb],
                  bloc.semaine[nsb] - bloc.periode, heure].groupe ;
            IF groupe >= 'X' THEN tableau[bloc.matiere[nsb],
                     bloc.semaine[nsb], heure].groupe := groupe
            ELSE
               BEGIN
                  heurp := heure MOD nb_heures + 1 ;
                  WHILE tableau[bloc.matiere[nsb], bloc.semaine[nsb] -
                        bloc.periode, heurp].groupe >= 'X' DO
                     heurp := heurp MOD nb_heures + 1 ;
                  tableau[bloc.matiere[nsb], bloc.semaine[nsb],
                     heure].groupe := tableau[bloc.matiere[nsb],
                        bloc.semaine[nsb] - bloc.periode, heurp].groupe ;
            END ;
            inc(heure) ;
         UNTIL heure > nb_heures ;
      END ;
 END ; { circulation_sousbloc }


 PROCEDURE remplis_matiere(mode_circulation : BOOLEAN) ;

 VAR     ix, semaine : WORD ;
         bloc : type_bloc ;
         matiere : CHAR ;

 BEGIN
   matiere := curseur.matiere ;
   semaine := curseur.semaine ;
   bloc := liste_blocs[recherche_bloc(matiere)] ;
   IF semaine <= bloc.periode THEN remplis_premiere(matiere) ;
   inc(semaine, bloc.periode - (semaine - 1) MOD bloc.periode) ;
   FOR ix := 1 TO bloc.long DO inc(bloc.semaine[ix], semaine) ;
   WHILE semaine <= nb_semaines DO
      BEGIN
         IF mode_circulation THEN circulation_sousbloc(bloc)
            ELSE circulation_bloc(bloc) ;
         FOR ix := 1 TO bloc.long DO inc(bloc.semaine[ix], bloc.periode) ;
         inc(semaine, bloc.periode) ;
      END ;
 END ; { remplis_matiere }


 PROCEDURE remplis_semaine(semaine : WORD) ;

 VAR     bloc : type_bloc ;
         matiere : CHAR ;
         ix, nsb : WORD ;

 BEGIN
   matiere := curseur.matiere ;
   FOR ix := 1 TO nb_blocs DO
      BEGIN
         bloc := liste_blocs[ix] ;
         IF (semaine - 1) MOD bloc.periode = 0 THEN
            BEGIN
               FOR nsb := 1 TO bloc.long DO inc(bloc.semaine[nsb], semaine) ;
               IF semaine > 1 THEN circulation_sousbloc(bloc)
               ELSE IF pos(matiere, bloc.matiere) > 0
                     THEN remplis_premiere(matiere)
                     ELSE remplis_premiere(bloc.matiere[1]) ;
            END ;
      END ;
 END ; { remplis_semaine }


 PROCEDURE remplis_optimise(semaine : WORD) ;

 VAR     numero, nsb, num_bloc, semainp : WORD ;
         bloc : type_bloc ;

 BEGIN
   semaine := semaine - (semaine - 1) MOD max_periode ;
   numero := 1 ;
   WHILE numero <= length(sequence_optimisation) DO
      BEGIN
         num_bloc := ord(sequence_optimisation[numero]) - ord('0') ;
         bloc := liste_blocs[num_bloc] ;
         FOR nsb := 1 TO bloc.long DO inc(bloc.semaine[nsb], semaine) ;
         semainp := semaine ;
         REPEAT
            IF (semainp - 1) MOD bloc.periode = 0 THEN
               BEGIN
                  IF semainp > 1 THEN circulation_sousbloc(bloc)
                     ELSE remplis_premiere(bloc.matiere[1]) ;
                  FOR nsb := 1 TO bloc.long DO
                     inc(bloc.semaine[nsb], bloc.periode) ;
               END ;
            inc(semainp) ;
         UNTIL semainp >= semaine + max_periode ;
         inc(numero) ;
      END ;
 END ; { remplis_optimise }


 PROCEDURE remplis_tableau ;

 VAR     semaine : WORD ;

 BEGIN
   FOR semaine := curseur.semaine TO nb_semaines DO remplis_semaine(semaine) ;
 END ; { remplis_tableau }

 {***************************************************************************
             Ralisation du colloscope - Fonctions d'optimisation
 ***************************************************************************}

 PROCEDURE tri_tds ;

 VAR     semaine : WORD ;
         matiere : CHAR ;

 PROCEDURE _tri ;

 VAR     heure, heurp : WORD ;
         temp : data ;

 BEGIN
   FOR heure := 1 TO nb_heures DO
      IF tableau[matiere, semaine, heure].groupe = 'X' THEN
         tableau[matiere, semaine, heure].groupe := 'x' ;
   FOR heure := 1 TO nb_heures - 1 DO                           { Tri bulle }
      FOR heurp := heure + 1 TO nb_heures DO
         IF tableau[matiere, semaine, heure].groupe >
            tableau[matiere, semaine, heurp].groupe THEN
      BEGIN
         temp := tableau[matiere, semaine, heurp] ;
         tableau[matiere, semaine, heurp] :=
               tableau[matiere, semaine, heure] ;
         tableau[matiere, semaine, heure] := temp ;
      END ;
   FOR heure := 1 TO nb_heures DO
      IF tableau[matiere, semaine, heure].groupe = 'x' THEN
         tableau[matiere, semaine, heure].groupe := 'X' ;
 END ; { _tri }

 BEGIN
   FOR semaine := 1 TO nb_semaines DO
         FOR matiere := 'a' TO derniere_matiere DO
      IF pos(matiere, liste_colles) = 0 THEN _tri ;
 END ; { tri_tds }


 PROCEDURE tri_initial(VAR initial : groupes_bloc ; limite : WORD) ;

 VAR     groupe, ixe : WORD ;

 BEGIN
   ixe := 0 ;
   REPEAT
      REPEAT
         inc(ixe) ;
      UNTIL initial[ixe] = 'X' ;
      groupe := ixe ;
      REPEAT
         inc(groupe) ;
      UNTIL (groupe > limite) OR (initial[groupe] < 'X') ;
      IF groupe <= limite THEN
         BEGIN
            initial[ixe] := initial[groupe] ;
            initial[groupe] := 'X' ;
         END ;
   UNTIL (groupe > limite) ;
 END ; { tri_initial }


 FUNCTION optimise(num_bloc, semaine : WORD) : WORD ;

 VAR     sous_ensemble : ARRAY[1..permutation_size] OF WORD ;
         etat_err : ARRAY[1..40] OF BOOLEAN ;   { erreur = false }
         initial, image : groupes_bloc ;
         nberrs_min, nb_errs : WORD ;
         reduit : BOOLEAN ;
         bloc : type_bloc ;

 PROCEDURE _cree_sous_ensemble ;

 VAR     ix, iy, iz, temp : WORD ;

 BEGIN
   iz := 1 ;
   ix := 1 ;
   WHILE (ix <= 2 * nb_heures) AND (iz < permutation_size) DO
      BEGIN
         IF NOT etat_err[ix] THEN
            BEGIN
               sous_ensemble[iz] := ix ;
               inc(iz) ;
            END ;
         inc(ix) ;
      END ;
   FOR ix := 1 TO 2 * nb_heures DO etat_err[ix] := true ;
   FOR ix := 1 TO iz - 1 DO etat_err[sous_ensemble[ix]] := false ;
   FOR ix := iz TO permutation_size DO
      BEGIN
         REPEAT
            iy := random(nb_elements) + 1 ;
         UNTIL etat_err[iy] ;
         sous_ensemble[ix] := iy ;
         etat_err[iy] := false ;
      END ;
   FOR ix := 1 TO permutation_size - 1 DO
      FOR iy := ix + 1 TO permutation_size DO
         IF sous_ensemble[ix] > sous_ensemble[iy] THEN
      BEGIN
         temp := sous_ensemble[ix] ;
         sous_ensemble[ix] := sous_ensemble[iy] ;
         sous_ensemble[iy] := temp ;
      END ;
   nb_elements := permutation_size ;
   reduit := true ;
 END ; { _cree_sous_ensemble }

 PROCEDURE _nouvel_essai ;

 VAR     heure, idx, ngp, nsb : WORD ;

 BEGIN
   idx := 1 ;
   ngp := 1 ;
   nsb := 1 ;
   IF reduit THEN
   REPEAT
      heure := 1 ;
      REPEAT
         IF image[heure + (nsb - 1) * nb_heures] < 'X' THEN
            BEGIN
               IF ngp = sous_ensemble[idx] THEN
                  BEGIN
                     tableau[bloc.matiere[nsb], bloc.semaine[nsb], heure].
                           groupe := initial[sous_ensemble[position[idx]]] ;
                     inc(idx) ;
                  END ;
               inc(ngp) ;
            END ;
         inc(heure) ;
      UNTIL (heure > nb_heures) OR (ngp > sous_ensemble[permutation_size]) ;
      inc(nsb) ;
   UNTIL (nsb > bloc.long) OR (ngp > sous_ensemble[permutation_size])
   ELSE
   REPEAT
      heure := 1 ;
      REPEAT
         IF image[heure + (nsb - 1) * nb_heures] < 'X' THEN
            BEGIN
               tableau[bloc.matiere[nsb], bloc.semaine[nsb], heure].groupe :=
                  initial[position[idx]] ;
               inc(idx) ;
            END ;
         inc(heure) ;
      UNTIL heure > nb_heures ;
      inc(nsb) ;
   UNTIL nsb > bloc.long ;
 END ; { _nouvel_essai }

 PROCEDURE _enregistre_essai ;

 VAR     heure, nsb : WORD ;

 BEGIN
   nberrs_min := nb_errs ;
   FOR nsb := 1 TO bloc.long DO
         FOR heure := 1 TO nb_heures DO
      image[heure + (nsb - 1) * nb_heures] :=
         tableau[bloc.matiere[nsb], bloc.semaine[nsb], heure].groupe ;
 END ; { _enregistre_essai }

 PROCEDURE _restaure_essai ;

 VAR     heure, nsb : WORD ;

 BEGIN
   FOR nsb := 1 TO bloc.long DO
         FOR heure := 1 TO nb_heures DO
      tableau[bloc.matiere[nsb], bloc.semaine[nsb], heure].groupe :=
         image[heure + (nsb - 1) * nb_heures] ;
 END ; { _restaure_essai }

 PROCEDURE _optimise_colles ;                     { permutation des groupes }

 VAR     heure, nsb : WORD ;
         groupe : CHAR ;

 BEGIN
   reduit := false ;
   nb_elements := 0 ;
   FOR heure := 1 TO 2 * nb_heures DO
      BEGIN
         etat_err[heure] := true ;
         position[heure] := heure ;
      END ;
   FOR nsb := 1 TO bloc.long DO
         FOR heure := 1 TO nb_heures DO
      BEGIN
         groupe := tableau[bloc.matiere[nsb],
                           bloc.semaine[nsb], heure].groupe ;
         initial[heure + (nsb - 1) * nb_heures] := groupe ;
         IF groupe < 'X' THEN
            BEGIN
               inc(nb_elements) ;
               etat_err[nb_elements] := (tableau[bloc.matiere[nsb],
                     bloc.semaine[nsb], heure].erreur = 0) ;
            END ;
      END ;
   image := initial ;
   tri_initial(initial, bloc.long * nb_heures) ;
   IF nb_elements > permutation_size THEN _cree_sous_ensemble ;
   nberrs_min := 9999 ;
   REPEAT
      _nouvel_essai ;
      FOR nsb := 1 TO bloc.periode DO test_optimisation(bloc.semaine[nsb]) ;
      nb_errs := decompte_erreurs(semaine) ;
      IF nb_errs <= nberrs_min THEN _enregistre_essai ;
      next_permutation ;
   UNTIL (position[1] = 0) ;
   _restaure_essai ;
 END ; { _optimise_colles }

 VAR     ix : WORD ;

 BEGIN
   modified := true ;
   bloc := liste_blocs[num_bloc] ;
   semaine := semaine - (semaine - 1) MOD bloc.periode ;
   FOR ix := 1 TO 4 * nb_heures DO initial[ix] := 'X' ;
   FOR ix := 1 TO bloc.long DO inc(bloc.semaine[ix], semaine) ;
   nb_errs := decompte_erreurs(semaine) ;
   IF nb_errs = 0 THEN nberrs_min := 0
      ELSE _optimise_colles ;
   evaluation_reduite(semaine) ;
   optimise := nberrs_min ;
 END ; { optimise }


 PROCEDURE affiche(reduit : BOOLEAN) ; FORWARD ;


 PROCEDURE optimisation_quinzaine ;

 VAR     numero, num_bloc, nb_errs, semaine : WORD ;

 BEGIN
   working := true ;
   semaine := curseur.semaine ;
   numero := random(length(sequence_optimisation)) + 1 ;
   REPEAT
      num_bloc := ord(sequence_optimisation[numero]) - ord('0') ;
      nb_errs := optimise(num_bloc, semaine) ;
      numero := numero MOD length(sequence_optimisation) + 1 ;
      IF numero = 1 THEN affiche(true) ;
   UNTIL keypressed OR (nb_errs = 0)  ;
   working := false ;
   tri_tds ;
   WHILE keypressed DO readkey ;
   beep ;
 END ; { optimisation_quinzaine }


 PROCEDURE optimisation_globale ;

 VAR     numero, num_bloc, passage, semaine, semaine_ref : WORD ;
         nb_errs : ARRAY[1..max_semaines] OF WORD ;
         initialise, stop, tested : BOOLEAN ;

 BEGIN
   stop := false ;
   working := true ;
   chi2_limite := false ;
   affiche(true) ;
   initialise := true ;
   FOR semaine := 1 TO nb_semaines DO
         nb_errs[semaine] := decompte_erreurs(semaine) ;
   REPEAT
      semaine := 1 ;
      IF initialise THEN passage := 0
         ELSE inc(passage) ;
      initialise := false ;
      REPEAT
         IF (passage > limite_optimisation) AND (nb_errs[semaine] > 0) THEN
            BEGIN
               remplis_optimise(semaine) ;
               initialise := true ;
            END ;
         semaine_ref := semaine - (semaine - 1) MOD max_periode ;
         tested := false ;
         numero := 1 ;
         WHILE (nb_errs[semaine] > 0) AND (NOT stop) AND
               (numero <= length(sequence_optimisation)) DO
            BEGIN
               num_bloc := ord(sequence_optimisation[numero]) - ord('0') ;
               IF (liste_blocs[num_bloc].periode = 1) OR
                     (semaine MOD liste_blocs[num_bloc].periode = 1) THEN
                  BEGIN
                     tested := true ;
                     nb_errs[semaine] := optimise(num_bloc, semaine) ;
                     IF (nb_errs[semaine] = 0) THEN
                        BEGIN
                           inc(duree_optimale, passage) ;
                           inc(nb_reussites) ;
                           IF nb_reussites = 4 THEN
                              limite_optimisation := duree_optimale ;
                        END ;
                  END ;
               inc(numero) ;
            END ;
         IF NOT tested THEN nb_errs[semaine] := 0 ;
         inc(semaine) ;
         stop := keypressed ;
      UNTIL stop OR (nb_erreurs = 0) OR (semaine > nb_semaines) ;
      evalue ;
      affiche(true) ;
   UNTIL stop OR (nb_erreurs = 0) ;
   working := false ;
   tri_tds ;
   WHILE keypressed DO readkey ;
   beep ;
 END ; { optimisation_globale }

 {***************************************************************************
             Ralisation du colloscope - Fonctions d'optimisation
              pour garantir la meilleure permutation des groupes
 ***************************************************************************}

 PROCEDURE chi2_global ;

 VAR     heure, limite, semaine : WORD ;
         groupe, matiere : CHAR ;

 BEGIN
   chi2_cumul := 0 ;
   IF chi2_reduit THEN limite := semaines_effectives
      ELSE limite := nb_semaines ;
   FOR matiere := 'a' TO derniere_matiere DO
         IF pos(matiere, liste_dispersion) > 0 THEN
      BEGIN
         FOR heure := 1 TO nb_heures DO
            FOR groupe := 'A' TO 'Z' DO data_chi2[heure, groupe] := 0 ;
         FOR heure := 1 TO nb_heures DO
               FOR semaine := 1 TO limite DO
            inc(data_chi2[heure, tableau[matiere, semaine, heure].groupe]) ;
         FOR heure := 1 TO nb_heures DO
               FOR groupe := 'A' TO 'W' DO
            inc(chi2_cumul, sqr(data_chi2[heure, groupe])) ;
      END ; { dispersion des passages }
   FOR matiere := 'a' TO derniere_matiere DO
         IF pos(matiere, liste_regular) > 0 THEN
      BEGIN                       { semaine = 1 + (semaine MOD max_periode) }
         FOR semaine := 1 TO max_periode DO
            FOR groupe := 'A' TO 'Z' DO data_chi2[semaine, groupe] := 0 ;
         FOR heure := 1 TO nb_heures DO
               FOR semaine := 1 TO limite DO
            inc(data_chi2[1 + (semaine MOD max_periode),
               tableau[matiere, semaine, heure].groupe]) ;
         FOR semaine := 1 TO max_periode DO
               FOR groupe := 'A' TO 'W' DO
            dec(chi2_cumul, sqr(data_chi2[semaine, groupe])) ;
      END ; { regularit des intervalles entre semaines de passage }
 END ; { chi2_global }


 PROCEDURE optimisation_chi2 ;

 VAR     chi2_dis, chi2_dis_min, chi2_reg, chi2_reg_max : INTEGER ;
         nbers, semaine : WORD ;
         matiere : CHAR ;

 PROCEDURE _init_data_dis ;

 VAR     heure, semip : WORD ;
         groupe : CHAR ;

 BEGIN
   FOR heure := 1 TO nb_heures DO
      FOR groupe := 'A' TO 'Z' DO data_chi2[heure, groupe] := 0 ;
   FOR heure := 1 TO nb_heures DO
         FOR semip := 1 TO nb_semaines DO
      IF (semip <> semaine) AND (tableau[matiere, semip, heure].groupe < 'X')
         THEN inc(data_chi2[heure, tableau[matiere, semip, heure].groupe]) ;
 END ; { _init_data_dis }

 PROCEDURE _optimise_chi2_dis ;

 VAR     nbers_min, heure, index : WORD ;
         image, initial : groupes_bloc ;

 BEGIN
   nb_elements := 0 ;
   FOR heure := 1 TO nb_heures DO
      BEGIN
         position[heure] := heure ;
         initial[heure] := tableau[matiere, semaine, heure].groupe ;
         IF initial[heure] < 'X' THEN inc(nb_elements) ;
      END ;
   FOR heure := nb_heures + 1 TO 4 * nb_heures DO initial[heure] := 'X' ;
   IF nb_elements = 0 THEN exit ;
   _init_data_dis ;
   tri_initial(initial, nb_heures) ;
   chi2_dis_min := 32760 ;
   nbers_min := 9999 ;
   REPEAT
      index := 1 ;
      FOR heure := 1 TO nb_heures DO
            IF tableau[matiere, semaine, heure].groupe < 'X' THEN
         BEGIN
            tableau[matiere, semaine, heure].groupe :=
                     initial[position[index]] ;
            inc(index) ;
         END ;
      test_optimisation(semaine) ;
      nbers := decompte_erreurs(semaine) ;
      chi2_dis := 0 ;
      FOR heure := 1 TO nb_heures DO inc(chi2_dis, data_chi2[heure,
            tableau[matiere, semaine, heure].groupe]) ;
      IF (nbers <= nbers_min) AND (chi2_dis < chi2_dis_min) THEN
         BEGIN
            nbers_min := nbers ;
            chi2_dis_min := chi2_dis ;
            FOR heure := 1 TO nb_heures DO
                  image[heure] := tableau[matiere, semaine, heure].groupe ;
         END ;
      next_permutation ;
   UNTIL (position[1] = 0) ;
   FOR heure := 1 TO nb_heures DO
         tableau[matiere, semaine, heure].groupe := image[heure] ;
 END ; { _optimise_chi2_dis }

 PROCEDURE _init_data_reg ;

 VAR     heure, semip : WORD ;
         groupe : CHAR ;

 BEGIN
   FOR semip := 1 TO max_periode DO
      FOR groupe := 'A' TO 'Z' DO data_chi2[semip, groupe] := 0 ;
   FOR heure := 1 TO nb_heures DO
         FOR semip := 1 TO nb_semaines DO
      IF (semip <> semaine) AND (tableau[matiere, semip, heure].groupe < 'X')
         THEN inc(data_chi2[1 + semip MOD max_periode,
            tableau[matiere, semip, heure].groupe]) ;
 END ; { _init_data_reg }

 PROCEDURE _optimise_chi2_reg ;

 VAR     nbers_min, heure, index, semaine_mod : WORD ;
         image, initial : groupes_bloc ;

 BEGIN
   nb_elements := 0 ;
   FOR heure := 1 TO nb_heures DO
      BEGIN
         position[heure] := heure ;
         initial[heure] := tableau[matiere, semaine, heure].groupe ;
         IF initial[heure] < 'X' THEN inc(nb_elements) ;
      END ;
   FOR heure := nb_heures + 1 TO 4 * nb_heures DO initial[heure] := 'X' ;
   IF nb_elements = 0 THEN exit ;
   _init_data_reg ;
   tri_initial(initial, nb_heures) ;
   chi2_reg_max := -32760 ;
   nbers_min := 9999 ;
   semaine_mod := 1 + semaine MOD max_periode ;
   REPEAT
      index := 1 ;
      FOR heure := 1 TO nb_heures DO
            IF tableau[matiere, semaine, heure].groupe < 'X' THEN
         BEGIN
            tableau[matiere, semaine, heure].groupe :=
                     initial[position[index]] ;
            inc(index) ;
         END ;
      test_optimisation(semaine) ;
      nbers := decompte_erreurs(semaine) ;
      chi2_reg := 0 ;
      FOR heure := 1 TO nb_heures DO inc(chi2_reg, data_chi2[semaine_mod,
            tableau[matiere, semaine, heure].groupe]) ;
      IF (nbers <= nbers_min) AND (chi2_reg > chi2_reg_max) THEN
         BEGIN
            nbers_min := nbers ;
            chi2_reg_max := chi2_reg ;
            FOR heure := 1 TO nb_heures DO
                  image[heure] := tableau[matiere, semaine, heure].groupe ;
         END ;
      next_permutation ;
   UNTIL (position[1] = 0) ;
   FOR heure := 1 TO nb_heures DO
         tableau[matiere, semaine, heure].groupe := image[heure] ;
 END ; { _optimise_chi2_reg }

 VAR     old_chi2 : INTEGER ;
         stop : BOOLEAN ;

 BEGIN
   chi2_reduit := false ;
   stop := false ;
   chi2_global ;
   affiche(false) ;
   REPEAT
      old_chi2 := chi2_cumul ;
      FOR matiere := 'a' TO derniere_matiere DO
            IF (pos(matiere, liste_colles) > 0) AND NOT stop THEN
         BEGIN
            semaine := 1 ;
            REPEAT
               IF (pos(matiere, liste_dispersion) > 0) THEN
                  _optimise_chi2_dis ;
               IF (pos(matiere, liste_regular) > 0) THEN
                  _optimise_chi2_reg ;
               evalue ;
               chi2_global ;
               affiche(true) ;
               inc(semaine) ;
               stop := keypressed ;
            UNTIL (semaine > nb_semaines) OR stop ;
         END ;
   UNTIL (abs(old_chi2 - chi2_cumul) < 5) OR stop ;
   working := false ;
   evalue ;
   WHILE keypressed DO readkey ;
   chi2_reduit := true ;
 END ; { optimisation_chi2 }


 FUNCTION similitude(cycle, cyclp : WORD) : WORD ;

 VAR     ix, iy, heure, semaine, semainp, semainx, semainy, simili : WORD ;
         matiere : CHAR ;

 BEGIN
   similitude := 0 ;
   IF cycle = cyclp THEN exit ;
   IF color[cycle * max_periode] OR color[cyclp * max_periode] THEN
      BEGIN
         similitude := 99 ;
         exit ;
      END ;
   simili := 0 ;
   semaine := (cycle - 1) * max_periode ;
   semainp := (cyclp - 1) * max_periode ;
   FOR ix := 1 TO max_periode DO
         FOR iy := 1 TO max_periode DO
      BEGIN
         semainx := semaine + ix ;
         semainy := semainp + iy ;
         FOR matiere := 'a' TO derniere_matiere DO
            IF pos(matiere, liste_colles) > 0 THEN
               FOR heure := 1 TO nb_heures DO
            BEGIN
               IF (tableau[matiere, semainx, heure].groupe =
                     tableau[matiere, semainy, heure].groupe)
                  AND (tableau[matiere, semainx, heure].groupe < 'X') THEN
                     inc(simili) ;
            END ;
      END ;
   similitude := simili ;
 END ; { similitude }


 PROCEDURE permute_cycles(cycle, cyclp : WORD) ;

 VAR     heure, semaine, semainp : WORD ;
         matiere : CHAR ;
         temp : data ;

 BEGIN
   IF cycle <> cyclp THEN
   FOR semaine := (cycle - 1) * max_periode + 1 TO cycle * max_periode DO
      BEGIN
         semainp := semaine + (cyclp - cycle) * max_periode ;
         FOR matiere := 'a' TO derniere_matiere DO
               FOR heure := 1 TO nb_heures DO
            BEGIN
               temp := tableau[matiere, semainp, heure] ;
               tableau[matiere, semainp, heure] :=
                     tableau[matiere, semaine, heure] ;
               tableau[matiere, semaine, heure] := temp ;
            END ;
      END ;
 END ; { permute_cycles }


 PROCEDURE calcule_matrice ;

 VAR     cycle, cyclp : WORD ;

 BEGIN
   evalue ;
   FOR cycle := 1 TO cycle_max DO
      FOR cyclp := 1 TO cycle_max DO
         matrice[cycle, cyclp] := similitude(cycle, cyclp) ;
 END ; { calcule_matrice }


 PROCEDURE calcule_resultat ;

 VAR     cycle, cyclp, somme : WORD ;

 BEGIN
   somme := 0 ;
   FOR cycle := 1 TO cycle_min DO
      FOR cyclp := cycle + 1 TO cycle_eff DO
         inc(somme, matrice[cycle, cyclp]) ;
   nb_similitudes := 2 * somme / (cycle_eff * cycle_min) ;
   sous_diagonale := 0 ;
   FOR cycle := 1 TO cycle_min DO
         inc(sous_diagonale, matrice[cycle, cycle + 1]) ;
 END ; { calcule_resultat }


 PROCEDURE test_similitude ;

 VAR     resultat : ARRAY[1..max_semaines] OF WORD ;
         limite : LONGINT ;

 PROCEDURE _optimise_sousdiagonale ;

 VAR     cycle, cyclp, temp, valeur, valeur_min : WORD ;
         fact : EXTENDED ;
         test : BOOLEAN ;

 BEGIN
   FOR cycle := 1 TO cycle_eff DO position[cycle] := cycle ;
   fact := ln_fact(cycle_eff) ;
   test := (fact > 16) ;                                   { cycle_eff > 10 }
   IF test THEN limite := $400000                      { 10! < limite < 11! }
      ELSE limite := round(exp(fact)) ;
   valeur_min := 9999 ;
   REPEAT
      valeur := 0 ;
      FOR cycle := 1 TO cycle_min DO
            inc(valeur, matrice[position[cycle], position[cycle + 1]]) ;
      IF valeur < valeur_min THEN
         BEGIN
            FOR cycle := 1 TO cycle_eff DO
                  resultat[cycle] := position[cycle] ;
            valeur_min := valeur ;
         END ;
      IF test THEN new_permutation                    { recherche partielle }
         ELSE next_permutation ;                    { ou recherche complte }
      dec(limite) ;
   UNTIL limite = 0 ;
   sous_diagonale := valeur_min ;
   FOR cycle := 1 TO cycle_eff DO position[resultat[cycle]] := cycle ;
   FOR cycle := 1 TO cycle_min DO
      FOR cyclp := cycle + 1 TO cycle_eff DO
         IF position[cycle] > position[cyclp] THEN
            BEGIN
               permute_cycles(cycle, cyclp) ;
               temp := position[cycle] ;
               position[cycle] := position[cyclp] ;
               position[cyclp] := temp ;
            END ;
 END ; { _optimise_sousdiagonale }

 PROCEDURE _tri_matrice ;

 VAR     cycle, cyclp, valeur, valeur_min : WORD ;

 BEGIN
   valeur_min := 9999 ;
   nb_elements := cycle_eff ;
   FOR cycle := 1 TO cycle_eff DO position[cycle] := cycle ;
   REPEAT
      valeur := 0 ;
      FOR cycle := 1 TO cycle_min DO
         FOR cyclp := cycle + 1 TO cycle_eff DO
            inc(valeur, matrice[position[cycle], position[cyclp]]) ;
      IF valeur < valeur_min THEN
         BEGIN
            FOR cycle := 1 TO cycle_eff DO
                  resultat[cycle] := position[cycle] ;
            valeur_min := valeur ;
         END ;
      next_combinaison ;
   UNTIL position[cycle_eff] * max_periode > nb_semaines ;
   FOR cycle := 1 TO cycle_eff DO permute_cycles(cycle, resultat[cycle]) ;
 END ; { _tri_matrice }

 BEGIN
   calcule_matrice ;
   _tri_matrice ;
   calcule_matrice ;
   _optimise_sousdiagonale ;
   calcule_resultat ;
   evalue ;
 END ; { test_similitude }


 PROCEDURE disperse_data ;

 BEGIN
   cursor(lineset_off) ;
   working := true ;
   modified := true ;
   chi2_limite := true ;
   optimisation_chi2 ;
   beep ;
   test_similitude ;
   beep ;
   working := false ;
 END ; { disperse }

 {***************************************************************************
               Ralisation du colloscope - Gestion des donnes
 ***************************************************************************}

 PROCEDURE creation_blocs ;

 VAR     ix : WORD ;

 BEGIN
   max_periode := 1 ;
   FOR ix := 1 TO nb_blocs DO
      BEGIN
         liste_blocs[ix].matiere := definition_blocs[ix] ;
         CASE length(definition_blocs[ix]) OF
            1 : BEGIN
                  liste_blocs[ix].long := 1 ;
                  liste_blocs[ix].periode := 1 ;
                  liste_blocs[ix].semaine[1] := 0 ;
               END ;
            2 : BEGIN
                  IF definition_blocs[ix][1] <> definition_blocs[ix][2] THEN
                     BEGIN
                        liste_blocs[ix].long := 2 ;
                        liste_blocs[ix].periode := 1 ;
                        liste_blocs[ix].semaine[1] := 0 ;
                        liste_blocs[ix].semaine[2] := 0 ;
                     END
                  ELSE
                     BEGIN
                        liste_blocs[ix].long := 2 ;
                        liste_blocs[ix].periode := 2 ;
                        liste_blocs[ix].semaine[1] := 0 ;
                        liste_blocs[ix].semaine[2] := 1 ;
                        max_periode := ppcm(max_periode, 2) ;
                     END
               END ;
            3 : BEGIN
                  IF definition_blocs[ix][1] <> definition_blocs[ix][3] THEN
                     BEGIN
                        liste_blocs[ix].long := 3 ;
                        liste_blocs[ix].periode := 2 ;
                        liste_blocs[ix].semaine[1] := 0 ;
                        liste_blocs[ix].semaine[2] := 1 ;
                        liste_blocs[ix].semaine[3] := 0 ;
                        max_periode := ppcm(max_periode, 2) ;
                     END
                  ELSE
                     BEGIN
                        liste_blocs[ix].long := 3 ;
                        liste_blocs[ix].periode := 3 ;
                        liste_blocs[ix].semaine[1] := 0 ;
                        liste_blocs[ix].semaine[2] := 1 ;
                        liste_blocs[ix].semaine[3] := 2 ;
                        max_periode := ppcm(max_periode, 3) ;
                     END
               END ;
            4 : BEGIN
                  IF definition_blocs[ix][1] <> definition_blocs[ix][4] THEN
                     BEGIN
                        liste_blocs[ix].long := 4 ;
                        liste_blocs[ix].periode := 2 ;
                        liste_blocs[ix].semaine[1] := 0 ;
                        liste_blocs[ix].semaine[2] := 1 ;
                        liste_blocs[ix].semaine[3] := 0 ;
                        liste_blocs[ix].semaine[4] := 1 ;
                        max_periode := ppcm(max_periode, 2) ;
                     END
                  ELSE
                     BEGIN
                        liste_blocs[ix].long := 4 ;
                        liste_blocs[ix].periode := 4 ;
                        liste_blocs[ix].semaine[1] := 0 ;
                        liste_blocs[ix].semaine[2] := 1 ;
                        liste_blocs[ix].semaine[3] := 2 ;
                        liste_blocs[ix].semaine[4] := 3 ;
                        max_periode := ppcm(max_periode, 4) ;
                     END
               END ;
         END ;
      END ;
   cycle_eff := semaines_effectives DIV max_periode ;
   cycle_min := cycle_eff - 1 ;
   cycle_max := nb_semaines DIV max_periode ;
   min_periode := max_periode - 1 ;
   IF nb_semaines MOD max_periode > 0 THEN
         arret('Nombre de semaines incorrect') ;
 END ; { creation_blocs }


 PROCEDURE controle_input_data ;

 VAR     data, texte, fichier_controle : STRING ;
         control_file : TEXT ;
         erreur : INTEGER ;
 CONST
   message = '  traiter trop grand, modifiez le paramtre et recompilez' ;

 PROCEDURE _extraction ;

 BEGIN
   REPEAT
      REPEAT
         readln(control_file, texte) ;
      UNTIL (length(texte) > 0) OR eof(control_file) ;
      IF pos('{', texte) > 0 THEN
         texte := copy(texte, 1, pos('{', texte) - 1) ;
      REPEAT
         WHILE texte[length(texte)] = ' ' DO
            texte := copy(texte, 1, length(texte) - 1) ;
         WHILE texte[length(texte)] = #9 DO
            texte := copy(texte, 1, length(texte) - 1) ;
      UNTIL (texte[length(texte)] <> ' ') AND (texte[length(texte)] <> #9) ;
   UNTIL ((length(texte) > 0) AND (upcase(texte[2]) = texte[2]))
      OR eof(control_file) ;
   IF pos(' ', texte) = 0 THEN texte := texte + ' ' ;
 END ; { _extraction }

 FUNCTION _next : BOOLEAN ;

 BEGIN
   texte := copy(texte, pos(' ', texte) + 1, 255) ;
   WHILE pos(' ', texte) = 1 DO texte := copy(texte, 2, 255) ;
   IF pos(' ', texte) > 1 THEN data := copy(texte, 1, pos(' ', texte) - 1)
   ELSE
      BEGIN
         data := texte ;
         texte := '' ;
      END ;
   _next := length(data) > 0 ;
 END ; { _next }

 PROCEDURE _amenagements ;

 BEGIN
   amenagements := true ;
   WHILE _next DO
      BEGIN
         IF length(data) = 2 THEN
            BEGIN
               inc(nb_amenagements_sup) ;
               amenagements_sup[nb_amenagements_sup] := data ;
               IF exclusion(data) THEN
                     arret('Liste des amnagements colles incorrecte') ;
            END
         ELSE
            BEGIN
               inc(nb_amenagements_tds) ;
               amenagements_tds[nb_amenagements_tds] := data ;
               IF exclusion(data) THEN
                     arret('Liste des amnagements TDs incorrecte') ;
            END ;
      END ;
   IF nb_amenagements_sup > max_blocs
      THEN arret('Nombre d''amnagements' + message) ;
   IF nb_amenagements_tds > max_blocs
      THEN arret('Nombre d''amnagements' + message) ;
 END ; { _amenagements }

 PROCEDURE _blocs ;

 BEGIN
   WHILE _next DO
      BEGIN
         inc(nb_blocs) ;
         definition_blocs[nb_blocs] := data ;
         IF exclusion(data) THEN arret('Liste des blocs incorrecte') ;
      END ;
   IF nb_blocs > max_blocs THEN arret('Nombre de blocs' + message) ;
 END ; { _blocs }

 PROCEDURE _colles ;

 VAR     index : INTEGER ;

 BEGIN
   _next ;
   liste_colles := data ;
   inc(nb_matieres, length(liste_colles)) ;
   inc(derniere_matiere, length(liste_colles)) ;
   liste_symboles := liste_symboles + liste_colles ;
   IF nb_matieres > max_matieres THEN arret('Nombre de matires' + message) ;
   FOR index := 1 TO length(liste_colles) DO
      IF (ord(liste_colles[index]) - ord('a') >= nb_matieres) OR
         (ord(liste_colles[index]) < ord('a')) OR
         (pos(liste_colles[index], liste_colles) < index) OR
         (pos(liste_colles[index], liste_tds) > 0) THEN
            arret('Liste des colles incorrecte') ;
   nb_colles := length(liste_colles) ;
 END ; { _colles }

 PROCEDURE _collisions ;

 BEGIN
   WHILE _next DO
      BEGIN
         inc(nb_collisions_tds) ;
         IF nb_collisions_tds > max_conflits THEN
            arret('Nombre de collisions' + message) ;
         collisions_tds[nb_collisions_tds] := data ;
         IF (pos(data[1], liste_colles) = 0) OR
            (pos(data[3], liste_tds) = 0) OR exclusion(data) THEN
               arret('Liste des collisions incorrecte') ;
      END ;
   IF nb_collisions_tds > max_conflits THEN
         arret('Nombre de collisions' + message) ;
 END ; { _collisions }

 PROCEDURE _conflits ;

 BEGIN
   WHILE _next DO
      BEGIN
         inc(nb_conflits_colles) ;
         IF nb_conflits_colles > max_conflits THEN
            arret('Nombre de conflits' + message) ;
         conflits_colles[nb_conflits_colles] := data ;
         IF (pos(data[1], liste_colles) = 0) OR
            (pos(data[3], liste_colles) = 0) OR exclusion(data) THEN
               arret('Liste des conflits incorrecte') ;
      END ;
   IF nb_conflits_colles > max_conflits THEN
         arret('Nombre de conflits' + message) ;
 END ; { _conflits }

 PROCEDURE _contraintes ;

 BEGIN
   WHILE _next DO
      BEGIN
         inc(nb_contraintes_colles) ;
         IF nb_contraintes_colles > max_conflits THEN
            arret('Nombre de contraintes' + message) ;
         contraintes_colles[nb_contraintes_colles] := data ;
         IF exclusion(data) THEN arret('Liste des contraintes incorrecte') ;
      END ;
   IF nb_contraintes_colles > max_conflits THEN
         arret('Nombre de contraintes' + message) ;
 END ; { _contraintes }

 PROCEDURE _dispersion ;

 VAR     index : INTEGER ;

 BEGIN
   _next ;
   liste_dispersion := data ;
   FOR index := 1 TO length(liste_dispersion) DO
      IF (ord(liste_dispersion[index]) - ord('a') >= nb_colles) OR
         (ord(liste_dispersion[index]) < ord('a')) OR
         (pos(liste_dispersion[index], liste_dispersion) < index) OR
         (pos(liste_dispersion[index], liste_tds) > 0) THEN
            arret('Liste dispersion incorrecte') ;
 END ; { _dispersion }

 PROCEDURE _nombre_groupes ;

 VAR     index : WORD ;

 BEGIN
   _next ;
   val(data, nb_groupes, erreur) ;
   nb_heures := (nb_groupes + 1) SHR 1 ;
   inc(dernier_groupe, nb_groupes) ;
   IF nb_heures > max_heures THEN arret('Nombre de groupes' + message) ;
   FOR index := 1 TO nb_heures DO liste_symboles := liste_symboles +
         chr(ord('0') + index) ;
   FOR index := 1 TO nb_groupes DO liste_symboles := liste_symboles +
         chr(ord('A') + index - 1) ;
 END ; { _nombre_groupes }

 PROCEDURE _obligations ;

 BEGIN
   WHILE _next DO
      BEGIN
         inc(nb_obligations_langues) ;
         obligations_langues[nb_obligations_langues] := data ;
         IF (pos(data[2], liste_colles) = 0) OR exclusion(data) THEN
                  arret('Liste des obligations incorrecte') ;
      END ;
   IF nb_obligations_langues > max_conflits THEN
         arret('Nombre d''obligations' + message) ;
 END ; { _obligations }

 PROCEDURE _optimisation ;

 VAR     index : WORD ;

 BEGIN
   _next ;
   sequence_optimisation := data ;
   FOR index := 1 TO length(data) DO
      IF ord(data[index]) > nb_blocs + ord('0') THEN
            arret('Squence d''optimisation incorrecte') ;
 END ; { _optimisation }

 PROCEDURE _regularite ;

 VAR     index : INTEGER ;

 BEGIN
   _next ;
   liste_regular := data ;
   FOR index := 1 TO length(liste_regular) DO
      IF (ord(liste_regular[index]) - ord('a') >= nb_colles) OR
         (ord(liste_regular[index]) < ord('a')) OR
         (pos(liste_regular[index], liste_regular) < index) OR
         (pos(liste_regular[index], liste_tds) > 0) THEN
            arret('Liste rgularit incorrecte') ;
 END ; { _regularite }

 PROCEDURE _semaines_optimisees ;

 BEGIN
   _next ;
   val(data, nb_semaines, erreur) ;
   IF (nb_semaines < semaines_effectives) OR (nb_semaines > max_semaines)
      THEN arret('Nombre de semaines optimises incorrect') ;
 END ; { _semaines_optimisees }

 PROCEDURE _semaines_effectives ;

 BEGIN
   _next ;
   val(data, semaines_effectives, erreur) ;
   IF (nb_semaines > 0) AND (nb_semaines < semaines_effectives) THEN
         arret('Nombre de semaines effectives incorrect') ;
 END ; { _semaines_effectives }

 PROCEDURE _successions ;

 BEGIN
   WHILE _next DO
      BEGIN
         inc(nb_successions_colles) ;
         successions_colles[nb_successions_colles] := data ;
         IF exclusion(data) THEN arret('Liste des successions incorrecte') ;
      END ;
   IF nb_successions_colles > max_conflits THEN
         arret('Nombre de successions' + message) ;
 END ; { _successions }

 PROCEDURE _tds ;

 VAR     index: INTEGER ;

 BEGIN
   _next ;
   liste_tds := data ;
   inc(nb_matieres, length(data)) ;
   inc(derniere_matiere, length(data)) ;
   liste_symboles := liste_symboles + data ;
   IF nb_matieres > max_matieres THEN arret('Nombre de matires' + message) ;
   FOR index := 1 TO length(data) DO
      IF (ord(data[index]) - ord('a') >= nb_matieres) OR
         (ord(data[index]) < ord('a')) OR
         (pos(data[index], liste_tds) < index) OR
         (pos(data[index], liste_colles) > 0) THEN
            arret('Liste des TDs incorrecte') ;
 END ; { _tds }

 PROCEDURE _titre ;

 BEGIN
   texte := copy(texte, pos(' ', texte) + 1, 255) ;
   IF titres[1] = '' THEN titres[1] := texte
      ELSE titres[2] := texte
 END ; { _titre }

 BEGIN
   titres[1] := '' ;
   titres[2] := '' ;
   amenagements := false ;
   liste_symboles := '+-' ;
   dernier_groupe := pred('A') ;
   derniere_matiere := pred('a') ;
   erreur := 0 ;
   nb_blocs := 0 ;
   nb_matieres := 0 ;
   nb_collisions_tds := 0 ;
   nb_conflits_colles := 0 ;
   nb_amenagements_tds := 0 ;
   nb_amenagements_sup := 0 ;
   nb_contraintes_colles := 0 ;
   nb_successions_colles := 0 ;
   nb_obligations_langues := 0 ;
   IF paramcount <> 1 THEN arret('Entrez le nom du fichier de donnes' +
      ' sur la ligne de commande !') ;
   fichier_controle := paramstr(1) ;
   IF NOT existe(fichier_controle) THEN
      arret('Nom de fichier de donnes invalide : ' + fichier_controle) ;
   reference_file := copy(fichier_controle, 1,
      pos('.', fichier_controle) - 1) + '_dat' + copy(fichier_controle,
      pos('.', fichier_controle), 255) ;
   result_file := copy(fichier_controle, 1, pos('.', fichier_controle) - 1) +
      '_res' + copy(fichier_controle, pos('.', fichier_controle), 255) ;
   assign(control_file, fichier_controle) ;
   reset(control_file) ;
   REPEAT
      _extraction ;
      IF pos('TDS', texte) = 1                  THEN _tds ;
      IF pos('TITRE', texte) = 1                THEN _titre ;
      IF pos('BLOCS', texte) = 1                THEN _blocs ;
      IF pos('COLLES', texte) = 1               THEN _colles ;
      IF pos('CONFLITS', texte) = 1             THEN _conflits ;
      IF pos('COLLISIONS', texte) = 1           THEN _collisions ;
      IF pos('DISPERSION', texte) = 1           THEN _dispersion ;
      IF pos('REGULARITE', texte) = 1           THEN _regularite ;
      IF pos('CONTRAINTES', texte) = 1          THEN _contraintes ;
      IF pos('OBLIGATIONS', texte) = 1          THEN _obligations ;
      IF pos('SUCCESSIONS', texte) = 1          THEN _successions ;
      IF pos('AMENAGEMENTS', texte) = 1         THEN _amenagements ;
      IF pos('OPTIMISATION', texte) = 1         THEN _optimisation ;
      IF pos('NOMBRE_GROUPES', texte) = 1       THEN _nombre_groupes ;
      IF pos('SEMAINES_EFFECTIVES', texte) = 1  THEN _semaines_effectives ;
      IF pos('SEMAINES_OPTIMISEES', texte) = 1  THEN _semaines_optimisees ;
      IF erreur > 0 THEN
         arret('Erreur numrique dans le fichier de donnes') ;
   UNTIL eof(control_file) ;
   IF (nb_blocs = 0) OR (nb_matieres = 0) THEN
            arret('Corrigez le fichier de donnes : ' + fichier_controle) ;
 { IF (nb_blocs = 0) OR (nb_matieres = 0) OR (nb_collisions_tds = 0) OR
      (nb_conflits_colles = 0) OR (nb_contraintes_colles = 0) OR
         (nb_obligations_langues = 0) THEN
            arret('Corrigez le fichier de donnes : ' + fichier_controle) ;{}
   close(control_file) ;
 END ; { controle_input_data }


 PROCEDURE init ;

 VAR     num : WORD ;

 BEGIN
   clrscr ;
   randomize ;
   new(memoire_tab) ;
   working := false ;
   modified := false ;
   cursor(lineset_off) ;
   chi2_reduit := true ;
   chi2_limite := false ;
   mode_curseur := true ; { mouvement du curseur - dplacement aprs frappe }
   textbackground(white) ;
   mode_torique := false ;       { mouvement du curseur - gomtrie torique }
   FOR num := 0 TO max_buffer DO subs_group[num] := 'Z' ;
   FOR num := 1 TO max_blocs DO amenagements_sup[num] := 'ZZZZ' ;
   FOR num := 1 TO max_blocs DO amenagements_tds[num] := 'ZZZZ' ;
   FOR num := 1 TO max_blocs DO definition_blocs[num] := 'ZZZZ' ;
   FOR num := 1 TO max_conflits DO collisions_tds[num] := 'ZZZZ' ;
   FOR num := 1 TO max_conflits DO conflits_colles[num] := 'ZZZZ' ;
   FOR num := 1 TO max_conflits DO contraintes_colles[num] := 'ZZZZ' ;
   FOR num := 1 TO max_conflits DO successions_colles[num] := 'ZZZZ' ;
   FOR num := 1 TO max_conflits DO obligations_langues[num] := 'ZZZZ' ;
   premiere := 1 ;
   liste_tds := 'ZZZZ' ;
   nb_erreurs := 0 ;
   chi2_cumul := 0 ;
   result_file := 'ZZZZ' ;
   nb_reussites := 0 ;
   liste_colles := 'ZZZZ' ;
   curseur.heure := 1 ;
   liste_regular := 'ZZZZ' ;
   duree_optimale := 9 ;
   dernier_groupe := 'Z' ;
   reference_file := 'ZZZZ' ;
   liste_symboles := 'ZZZZ' ;
   curseur.semaine := 1 ;
   curseur.matiere := 'a' ;
   max_amenagements := 0 ;
   derniere_matiere := 'Z' ;
   liste_dispersion := 'ZZZZ' ;
   limite_optimisation := 222 ;
   sequence_optimisation := 'ZZZZ' ;
   controle_input_data ;
   creation_blocs ;
   clear_tableau ;
   nb_lignes := inf(nb_semaines, max_lignes) ;
 END ; { init }


 PROCEDURE enregistre ;

 VAR     groupe, matiere, minus, saut : CHAR ;
         hstr, sstr : STRING[2] ;
         heure, semaine : WORD ;
         texte : FILE OF CHAR ;
         moins : BOOLEAN ;

 PROCEDURE _new_line ;

 BEGIN
   saut := #$0d ;
   write(texte, saut) ;
   saut := #$0a ;
   write(texte, saut) ;
   saut := ' ' ;
 END ; { _new_line }

 PROCEDURE _write_string(ligne : STRING) ;

 VAR     ix : WORD ;

 BEGIN
   _new_line ;
   FOR ix := 1 TO length(ligne) DO write(texte, ligne[ix]) ;
   _new_line ;
   _new_line ;
 END ; { _write_string }

 BEGIN
   IF NOT modified THEN exit ;
   modified := false ;
   saut := ' ' ;
   minus := '-' ;

   assign(texte, result_file) ;
   rewrite(texte) ;
   _write_string('Classement Semaines \ Colles -> Groupes') ;
   write(texte, minus) ;
   FOR matiere := 'a' TO derniere_matiere DO
         FOR heure := 1 TO nb_heures DO
      BEGIN
         write(texte, saut) ;
         hstr := str_hex(heure, 0) ;
         write(texte, hstr[1]) ;
      END ;
   _new_line ;
   FOR semaine := 1 TO semaines_effectives DO
      BEGIN
         str(semaine, sstr) ;
         write(texte, sstr[1]) ;
         IF length(sstr) = 2 THEN write(texte, sstr[2]) ;
         write(texte, saut) ;
         FOR matiere := 'a' TO derniere_matiere DO
               FOR heure := 1 TO nb_heures DO
            BEGIN
               write(texte, tableau[matiere, semaine, heure].groupe) ;
               IF (matiere < derniere_matiere) OR (heure < nb_heures) THEN
                  write(texte, saut) ;
            END ;
         _new_line ;
      END ;

   _write_string('Classement Colles \ Semaines -> Groupes') ;
   write(texte, minus) ;
   FOR semaine := 1 TO semaines_effectives DO
      BEGIN
         write(texte, saut) ;
         str(semaine, sstr) ;
         write(texte, sstr[1]) ;
         IF length(sstr) = 2 THEN write(texte, sstr[2]) ;
      END ;
   _new_line ;
   FOR matiere := 'a' TO derniere_matiere DO
         FOR heure := 1 TO nb_heures DO
      BEGIN
         write(texte, matiere) ;
         IF pos(matiere, liste_colles) > 0 THEN
            BEGIN
               hstr := str_hex(heure, 0) ;
               write(texte, hstr[1]) ;
            END ;
         FOR semaine := 1 TO semaines_effectives DO
            BEGIN
               write(texte, saut) ;
               write(texte, tableau[matiere, semaine, heure].groupe) ;
            END ;
         _new_line ;
      END ;

   _write_string('Classement Semaines \ Groupes -> Colles') ;
   write(texte, minus) ;
   FOR groupe := 'A' TO dernier_groupe DO
      BEGIN
         write(texte, saut) ;
         write(texte, groupe) ;
      END ;
   _new_line ;
   FOR semaine := 1 TO semaines_effectives DO
      BEGIN
         str(semaine:0, sstr) ;
         write(texte, sstr[1]) ;
         IF semaine > 9 THEN write(texte, sstr[2]) ;
         FOR groupe := 'A' TO dernier_groupe DO
            BEGIN
               moins := true ;
               write(texte, saut) ;
               FOR matiere := 'a' TO derniere_matiere DO
                  FOR heure := 1 TO nb_heures DO
               IF tableau[matiere, semaine, heure].groupe = groupe THEN
                  BEGIN
                     IF pos(matiere, liste_colles) > 0 THEN
                        BEGIN
                           write(texte, matiere) ;
                           hstr := str_hex(heure, 0) ;
                           write(texte, hstr[1]) ;
                        END
                     ELSE IF moins THEN
                        BEGIN
                           moins := false ;
                           write(texte, minus) ;
                           write(texte, matiere) ;
                        END
                     ELSE write(texte, matiere)
                  END ;
            END ;
         _new_line ;
      END ;
   close(texte) ;

   assign(texte, reference_file) ;
   rewrite(texte) ;
   FOR semaine := 1 TO nb_semaines DO
      FOR matiere := 'a' TO derniere_matiere DO
         FOR heure := 1 TO nb_heures DO
      write(texte, tableau[matiere, semaine, heure].groupe) ;
   close(texte) ;
 END ; { enregistre }


 PROCEDURE restaure ;

 VAR     semaine, heure : WORD ;
         texte : FILE OF CHAR ;
         matiere : CHAR ;

 BEGIN
   IF NOT existe(reference_file) THEN exit ;
   modified := false ;
   assign(texte, reference_file) ;
   reset(texte) ;
   IF filesize(texte) = nb_semaines * (nb_matieres * nb_heures) THEN
      FOR semaine := 1 TO nb_semaines DO
         FOR matiere := 'a' TO derniere_matiere DO
            FOR heure := 1 TO nb_heures DO
         read(texte, tableau[matiere, semaine, heure].groupe) ;
   close(texte) ;
 END ; { restaure }

 {***************************************************************************
              Ralisation du colloscope - Gestion de l'affichage
 ***************************************************************************}

 PROCEDURE affiche(reduit : BOOLEAN) ;

 VAR     color, color_ref, wx, wy : WORD ;
         texte : STRING[30] ;

 PROCEDURE _affiche_tableau ;

 VAR     debut_x, debut_y, heure, ligne, semaine : WORD ;
         matiere : CHAR ;

 BEGIN
   wx := 80 ;
   wy := 1 ;
   debut_x := (78 - nb_matieres * (nb_heures + 1)) SHR 1 ;
   debut_y := (47 - nb_lignes - nb_lignes SHR 1) SHR 1 ;
   IF (debut_x < 2) OR (debut_x > 80) THEN debut_x := 2 ;
   IF (debut_y < 2) OR (debut_y > 20) THEN debut_y := 4 ;
   FOR ligne := 1 TO nb_lignes DO
      BEGIN
         semaine := ligne + premiere - 1 ;
         IF clscp.color[semaine] THEN color_ref := yellow
            ELSE color_ref := white ;
         gotoxy(debut_x, debut_y + ligne + (ligne - 1) SHR 1) ;
         textcolor(green) ;
         write(semaine:2, ' ':2) ;
         FOR matiere := 'a' TO derniere_matiere DO
            BEGIN
               FOR heure := 1 TO nb_heures DO
                  BEGIN
                     color := black ;
                     IF tableau[matiere, semaine, heure].erreur > 0
                        THEN color := red
                        ELSE color := color_ref ;
                     IF tableau[matiere, semaine, heure].groupe = 'X'
                        THEN color := blue ;
                     IF tableau[matiere, semaine, heure].groupe = 'Z'
                        THEN color := green ;
                     IF (NOT working) AND (matiere = curseur.matiere) AND
                        (heure = curseur.heure) AND
                        (semaine = curseur.semaine) THEN
                        BEGIN
                           IF color = white THEN color := black + blink
                              ELSE color := color + blink ;
                           wx := wherex ;
                           wy := wherey ;
                        END ;
                     textcolor(color) ;
                     write(tableau[matiere, semaine, heure].groupe) ;
                  END ;
               textcolor(color_ref) ;
               write(' ') ;
            END ;
      END ;
 END ; { _affiche_tableau }

 FUNCTION _nb_chiffres(x : INTEGER) : WORD ;

 VAR     y : WORD ;

 BEGIN
   IF abs(x) = 0 THEN y := 1
      ELSE y := trunc(1.000001 + ln(abs(x)) / ln(10)) ;
   IF x <= -1 THEN inc(y) ;
   _nb_chiffres := y ;
 END ;

 PROCEDURE _affiche_erreurs ;

 VAR     index, semaines_conflits : WORD ;
         groupe : CHAR ;

 BEGIN
   semaines_conflits := 0 ;
   FOR index := 1 TO nb_semaines DO
      IF clscp.color[index] THEN inc(semaines_conflits) ;
   index := _nb_chiffres(nb_erreurs) + 4 * nb_groupes ;
   inc(index, _nb_chiffres(INTEGER(nb_erreurs) - INTEGER(old_nber))) ;
   inc(index, _nb_chiffres(semaines_conflits)) ;
   index := 35 - (index SHR 1) ;
   IF (index < 1) OR (index > 30) THEN index := 1 ;
   gotoxy(index, 43) ;
   textcolor(red) ;
   write(' ':2, nb_erreurs, ' ':2, INTEGER(nb_erreurs) - INTEGER(old_nber)) ;
   write(' ':2, semaines_conflits, ' ':4) ;
   old_nber := nb_erreurs ;
   FOR groupe := 'A' TO dernier_groupe DO
      BEGIN
         textcolor(white) ;
         write(groupe) ;
         textcolor(red) ;
         write(str_hex(erreur_groupes[groupe], 2), ' ') ;
      END ;
   write('  ') ;
   texte := messages_erreur[tableau[curseur.matiere, curseur.semaine,
         curseur.heure].erreur] ;
   gotoxy(5, 49) ;
   textcolor(red) ;
   write(' ':2, nb_tests, ' ':4) ;
   IF (length(texte) = 0) OR chi2_limite THEN
      BEGIN
         textcolor(green) ;
         write(nb_similitudes:0:1, ' ':3, sous_diagonale, ' ':3) ;
         IF NOT chi2_reduit THEN textcolor(red) ;
         write(chi2_cumul, ' ':3) ;
         textcolor(green) ;
         write(limite_optimisation, ' ':2) ;
      END
   ELSE write(' ':2, texte, ' ':22 - length(texte)) ;
 END ; { _affiche_erreurs }

 PROCEDURE _affiche_parametres ;

 VAR     numero : WORD ;

 BEGIN
   gotoxy(43, 49) ;
   textcolor(white) ;
   FOR numero := 0 TO 6 DO write(' ', subs_group[numero]) ;
   IF amenagements THEN
      BEGIN
         gotoxy(65, 49) ;
         write(str_hex(max_amenagements, 0), '',
            str_hex(nb_amenagements_tds, 0), '+',
               str_hex(nb_amenagements_sup, 0)) ;
      END
   ELSE gotoxy(70, 49) ;
   IF mode_curseur THEN write('  Mobile') ELSE write('  Fixe  ') ;
 END ; { _affiche_parametres }

 PROCEDURE _affiche_titre ;

 BEGIN
   textcolor(blue) ;
   gotoxy(1, 2) ;
   write(titres[1], ret, titres[2]) ;
 END ; { _affiche_titre }

 PROCEDURE _affiche_mode_emploi ;

 BEGIN
   textcolor(blue) ;
   gotoxy(1, 45) ;
   write(' ':((80 - length(mode_emploi[0])) SHR 1), mode_emploi[0], ret,
         ' ':((80 - length(mode_emploi[1])) SHR 1), mode_emploi[1], ret,
         ' ':((80 - length(mode_emploi[2] + mode_emploi[3])) SHR 1),
         mode_emploi[2]) ;
   textcolor(red) ;
   write(mode_emploi[3]) ;
 END ; { _affiche_mode_emploi }

 BEGIN
   cursor(lineset_off) ;
   _affiche_tableau ;
   _affiche_erreurs ;
   IF NOT reduit THEN
      BEGIN
         _affiche_parametres ;
         _affiche_mode_emploi ;
         _affiche_titre ;
         gotoxy(wx, wy) ;
      END
   ELSE gotoxy(1, 1) ;
   cursor(lineset_on) ;
 END ; { affiche }

 {***************************************************************************
          Ralisation du colloscope - Gestion des commandes clavier
 ***************************************************************************}

 PROCEDURE dec_matiere ;

 BEGIN
   IF curseur.matiere > 'a' THEN dec(curseur.matiere)
      ELSE IF mode_torique THEN curseur.matiere := derniere_matiere ;
 END ; { dec_matiere }


 PROCEDURE inc_matiere ;

 BEGIN
   IF curseur.matiere < derniere_matiere THEN inc(curseur.matiere)
      ELSE IF mode_torique THEN curseur.matiere := 'a' ;
 END ; { inc_matiere }


 PROCEDURE dec_heure ;

 BEGIN
   IF curseur.heure > 1 THEN dec(curseur.heure)
   ELSE IF mode_torique OR (curseur.matiere > 'a') THEN
      BEGIN
         curseur.heure := nb_heures ;
         dec_matiere ;
      END ;
 END ; { dec_heure }


 PROCEDURE inc_heure ;

 BEGIN
   IF curseur.heure < nb_heures THEN inc(curseur.heure)
   ELSE IF mode_torique OR (curseur.matiere < derniere_matiere) THEN
      BEGIN
         curseur.heure := 1 ;
         inc_matiere ;
      END ;
 END ; { inc_heure }


 FUNCTION edite : BOOLEAN ;

 VAR     ck, cl : CHAR ;

 PROCEDURE _init_limite(limite : WORD) ;

 BEGIN
   nb_reussites := 0 ;
   duree_optimale := 9 ;
   limite_optimisation := limite ;
 END ; { _init_limite }

 PROCEDURE _commandes ;                                         { COMMANDES }

 VAR     num_bloc : WORD ;

 BEGIN
   CASE cl OF
         #059 : remplis_matiere(true) ;                                { F1 }
         #060 : remplis_matiere(false) ;                               { F2 }
         #061 : remplis_semaine(curseur.semaine) ;                     { F3 }
         #062 : remplis_tableau ;                                      { F4 }
         #063 : BEGIN                                                  { F5 }
                  num_bloc := 1 ;
                  WHILE pos(curseur.matiere, definition_blocs[num_bloc]) = 0
                     DO inc(num_bloc) ;
                  optimise(num_bloc, curseur.semaine) ;
                  beep ;
               END ;
         #064 : optimisation_quinzaine ;                               { F6 }
         #065 : optimisation_globale ;                                 { F7 }
         #066 : disperse_data ;                                        { F8 }
         #067 : memoire_tab^ := tableau ;                              { F9 }
         #068 : tableau := memoire_tab^ ;                             { F10 }
         #133 : tri_tds ;                                             { F11 }
         #104 : _init_limite(19) ;                                 { Alt-F1 }
         #105 : _init_limite(71) ;                                 { Alt-F2 }
         #106 : _init_limite(267) ;                                { Alt-F3 }
         #107 : _init_limite(999) ;                                { Alt-F4 }
         #108 : mode_torique := NOT mode_torique ;                 { Alt-F5 }
         #109 : mode_curseur := NOT mode_curseur ;                 { Alt-F6 }
         #111 : clear_tableau ;                                    { Alt-F8 }
         #112 : enregistre ;                                       { Alt-F9 }
         #113 : restaure ;                                         { Alt-Fa }
      END ;
 END ; { _commandes }

 PROCEDURE _mouvements ;                            { MOUVEMENTS DU CURSEUR }

 BEGIN
   CASE cl OF
         #071 : BEGIN                                                { Home }
                  curseur.heure := 1 ;
                  curseur.semaine := 1 ;
                  curseur.matiere := 'a' ;
               END ;
         #072 : BEGIN                                            { Up Arrow }
                  dec(curseur.semaine) ;
                  IF (curseur.semaine = 0) AND mode_torique
                     THEN curseur.semaine := nb_semaines ;
                  IF (curseur.semaine = 0) AND NOT mode_torique
                     THEN curseur.semaine := 1 ;
               END ;
         #073 : curseur.semaine := 1 ;                            { Page Up }
         #075 : dec_heure ;                                    { Left Arrow }
         #077 : inc_heure ;                                   { Right Arrow }
         #079 : BEGIN                                                 { Fin }
                  curseur.heure := nb_heures ;
                  curseur.semaine := nb_semaines ;
                  curseur.matiere := derniere_matiere ;
               END ;
         #080 : BEGIN                                          { Down Arrow }
                  inc(curseur.semaine) ;
                  IF (curseur.semaine > nb_semaines) AND mode_torique
                     THEN curseur.semaine := 1 ;
                  IF (curseur.semaine > nb_semaines) AND NOT mode_torique
                     THEN curseur.semaine := nb_semaines ;
               END ;
         #081 : curseur.semaine := nb_semaines ;                { Page Down }
         #115 : dec_matiere ;                             { Ctrl-Left Arrow }
         #116 : inc_matiere ;                            { Ctrl-Right Arrow }
      END ;
 END ; { _mouvements }

 PROCEDURE _deplacement_fenetre ;                { MOUVEMENTS DE LA FENETRE }

 BEGIN
   CASE cl OF
      #118 : WHILE premiere + min_periode <=               { Ctrl-Page Down }
               nb_semaines - nb_lignes DO inc(premiere, max_periode) ;
      #132 : premiere := 1 ;                                 { Ctrl-Page Up }
      #141 : IF premiere > 1 THEN dec(premiere, max_periode) ;
                                                            { Ctrl-Up Arrow }
      #145 : IF premiere + min_periode <= nb_semaines - nb_lignes THEN
               inc(premiere, max_periode) ;               { Ctrl-Down Arrow }
      END ;
 END ; { _deplacement_fenetre }

 PROCEDURE _controles ;

 VAR     num : WORD ;

 BEGIN
   CASE ck OF
      'A'..'Z' : BEGIN
                     FOR num := max_buffer DOWNTO 1 DO
                           subs_group[num] := subs_group[num - 1] ;
                     subs_group[0] := tableau[curseur.matiere,
                           curseur.semaine, curseur.heure].groupe ;
                     tableau[curseur.matiere, curseur.semaine,
                           curseur.heure].groupe := ck ;
                     IF mode_curseur THEN inc_heure ;
                     modified := true ;
                  END ;
      '+' : IF amenagements AND (max_amenagements < nb_amenagements_tds +
                  nb_amenagements_sup) THEN inc(max_amenagements) ;
      '-' : IF amenagements AND (max_amenagements > 0)
                  THEN dec(max_amenagements) ;
      ' ' : BEGIN
               tableau[curseur.matiere, curseur.semaine,
                     curseur.heure].groupe := subs_group[0] ;
               FOR num := 0 TO max_buffer - 1 DO
                     subs_group[num] := subs_group[num + 1] ;
               IF mode_curseur THEN inc_heure ;
               modified := true ;
            END ;
      esc : edite := true ;
      END ;
 END ; { _controles }

 BEGIN
   edite := false ;
   cl := #0 ;
   ck := upcase(readkey) ;
   IF ck = #0 THEN cl := readkey ;
   WHILE keypressed DO readkey ;
   _controles ;
   _commandes ;
   _mouvements ;
   _deplacement_fenetre ;
 END ; { edite }

 {***************************************************************************
               Ralisation du colloscope - Programme principal
 ***************************************************************************}

 BEGIN
   init ;
   restaure ;
   evalue ;
   chi2_global ;
   calcule_matrice ;
   calcule_resultat ;
   memoire_tab^ := tableau ;
   REPEAT
      evalue ;
      chi2_global ;
      calcule_matrice ;
      calcule_resultat ;
      affiche(false) ;
   UNTIL edite ;
   enregistre ;
 END. { main }

 {***************************************************************************
                       Ralisation du colloscope - Fin
 ***************************************************************************}
